;----------------------------------------------------------------------------
;
; (C)Copyright 1989 Gibson Research, Incorporated
; All rights reserved.
;
; Creation date: 10/6/89
; Author . . . : Michael Toutonghi
;
;----------------------------------------------------------------------------

;----------------------------------------------------------------------------
;
; This file provides object oriented extensions to assembly
; language through the use of macros.
;
; The macros provided here follow:
;
;	OBJECT		- begin an object definition
;	OBJECT_END	- end an object definition
;
;	ODB		- data declaration
;	ODW
;	ODD
;	ODF
;	ODP
;	ODQ
;	ODT
;
;	METHOD		- define a method for this object
;	INIT_EXIT	- define an automatic initialization and exit method
;	SET_OBJECT_TYPE - set a memory block to an object type (used by init)
;
;	USE_OBJECTS	- automatic inclusion of object files with the
;			  assembly time constant $IsIncluded set to TRUE
;
;	To declare an object:
;
;		OBJECT objname [,ancestorname [,EXTERNAL]]
;		  [ ODB | ODW | ODD | ODT | ... ] itemname value [,PUBLIC]]
;		  .
;		  .
;		  .
;
;		  [INIT_EXIT initname,exitname]
;
;		  [METHOD methodname [,VIRTUAL [,INLINE] ] ]
;		OBJECT_END
;
;	A data item is the equivalent of a structure element. A method
;	is a macro which invokes a procedure of the name:
;
;		_objname_methodname
;
;	Each method must have a procedure associated with it unless it
;	is inline. If inline, a macro must be defined with it's name
;	before it is invoked. The standard way to invoke a method is:
;
;		objname_methodname [ObjInstance[,axparm[,bxparm[,cxparm[,dxparm]]]]]
;
;	The offset of ObjInstance, if present, will be loaded into di
;	with an LEA instruction. The parameters, if present, will be put
;	into the listed registers. If the ObjInstance value is not
;	present or [DI], it won't be loaded. If the method is virtual,
;	there must be a valid DGROUP assume and si will be used. I have
;	been using ES as the default DGROUP register and DS as the
;	segment of all objects with excellent results.
;
;	If an ancestor is declared, this object will inherit all of the
;	data items and methods of it's ancestor and you can add from
;	there. Later, if you add a method or data item to an ancestor,
;	all of it's children aquire that method or item.
;
;	You can declare a method to be virtual which will allow it
;	to be redefined by a child and bound at run time. Children can
;	redefine non-virtual methods, but the binding will occur at
;	assemble time.
;
;	If a method is virtual and it is invoked from a procedure or
;	method previously assembled, the type of object passed to that
;	procedure or method will determine the actual code which gets
;	executed.
;
;
;
;
;----------------------------------------------------------------------------

;; definitions of code and data sizes
CodePtrSize	EQU 2			;; size of a code pointer for .COM file

;; equates for standard code and data segments
@CODE		EQU <_TEXT>		;; code segment (take out for auto segs)
@DATA		EQU <_DATA>		;; data seg	(take out for auto segs)
@DATA_GROUP	EQU DGROUP		;; data group

;; equates to access offsets within the above segments
@DATAOFS	EQU OFFSET DGROUP:	;; offset of data,
@CODEOFS	EQU OFFSET DGROUP:	;; offset of code,

$CurObjectOpen	= FALSE			;; keeps track of open objects
$CurStructOpen	= FALSE			;; keeps track of open structures
$CurDataItem	= 0			;; the number of current data items
$TempMethodDefined = FALSE		;; predefine it
$IsIncluded	= FALSE			;; this is set for included objects

$ODTPointerStruc STRUC
  $ODTOfs	dw ?			;; this offset is present in all objects
$ODTPointerStruc ENDS

;; This calls a pointer with either a near or far call depending on
;; the code pointer size
$CALLPTR MACRO PToCall
  IF (CodePtrSize EQ 2)
    ;; call this method
    call word ptr PToCall
  ELSE
    ;; call the method far
    call dword ptr PToCall
  ENDIF
ENDM

; Macro to define an object. Ancestor is the optional
; object to inherit data elements and methods from.
; If anything is passed in ExternalFlag, it is assumed that
; the data and code declarations for this object are
; external and in another linkable file.
OBJECT		MACRO ObjName,Ancestor,ExternalFlag

  IF $CurObjectOpen
    .err
    %out Error: nesting of objects or structures is not allowed
    EXITM
  ENDIF

  ;; an object is open, keep track for error checking
  $CurObjectOpen = TRUE			;; the object indicator
  $CurStructOpen = TRUE			;; and current structure
  $CurDataItem	 = 0			;; start with no data items

  ;; ******************************************************
  ;; RULES FOR METHOD INVOCATIONS *************************
  ;; ******************************************************
  ;; A method invocation takes the form:
  ;;   ObjectName_MethodName ObjectParam,AXParm,BXParm,CXParm,DXParm
  ;;
  ;; To call the object's ancestor method:
  ;;   ObjectName_MethodName_A ObjectParam,AXParm,BXParm,CXParm,DXParm
  ;;
  ;; Parameters:
  ;;	 ds = segment of ObjectParam unless the method is non
  ;;	    virtual and any parameters are absolutely
  ;;	    unnecessary
  ;;
  ;;	 some segment register must be in DGROUP with the proper
  ;;	 ASSUME active
  ;;
  ;;   ObjectParam = offset of object
  ;; Action:
  ;;   defined by method
  ;; Uses regs:
  ;;   di and si are used for the invocation
  ;;   other registers may be used by the method
  ;; ******************************************************

  @SEG_START MACRO SEGM
    &&SEGM SEGMENT
  ENDM
  @SEG_END MACRO SEGM
    &&SEGM ENDS
  ENDM

  ;; This macro simply loads parameters into registers
  $GET_OBJ_REG_PARMS MACRO ObjectParm,P1,P2,P3,P4
    IFNB <P1>
      IFDIFI <P1>,<ax>
	mov ax,P1
      ENDIF
    ENDIF
    IFNB <P2>
      IFDIFI <P2>,<bx>
	mov bx,P2
      ENDIF
    ENDIF
    IFNB <P3>
      IFDIFI <P3>,<cx>
	mov cx,P3
      ENDIF
    ENDIF
    IFNB <P4>
      IFDIFI <P4>,<dx>
	mov dx,P4
      ENDIF
    ENDIF
    IFNB <ObjectParm>
      IFDIFI <[di]>,<ObjectParm>
	lea di,ObjectParm		;; point to object with di
      ENDIF
    ENDIF
  ENDM

  ;; This macro declares a routine external and depending on
  ;; CodePtrSize it makes it near or far.
  $DECL_EXTRN_ROUT MACRO RoutName
    IF (CodePtrSize EQ 2)		;; if size 2, near
      EXTRN &&RoutName:Near
    ELSE
      EXTRN &&RoutName:Far		;; else far
    ENDIF
  ENDM

  ;; This macro takes an existing object and converts it to a specific
  ;; type of object. This macro does not allocate memory or initialize
  ;; any variables. It can be used to set a dynamically allocated
  ;; object to a certain type. In fact, an object instance which uses
  ;; virtual methods will not work properly if it was not either
  ;; declared as a static instance or initialized with this macro.
  SET_OBJECT_TYPE MACRO ObjInstance,ObjectType
    ;; set pointer to ODT
    mov ObjInstance.$ODTOfs,@DATAOFS $&&ObjectType&&_ODT
  ENDM

  ;; This macro makes a macro to access the current object's
  ;; ancestor's method of the same name
  MAKE_ANCESTOR_METHOD MACRO MethodName,CurrentVNum
    ;; See if we'll be making a virtual or non-virtual method
    IFDEF $&&MethodName&&_Ofs
      ;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
      ObjName&_&&MethodName&&_A MACRO ObjectParm,P1,P2,P3,P4
	$GET_OBJ_REG_PARMS <ObjectParm>,<P1>,<P2>,<P3>,<P4>
	mov si,[di].$ODTOfs		;; get pointer to ODT
	$CALLPTR <DGROUP:[si].((CurrentVNum SHL (CodePtrSize SHR 1))-(CodePtrSize-2))>
      ENDM
    ELSE
      ;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
      ObjName&_&&MethodName&&_A MACRO ObjectParm,P1,P2,P3,P4
	Ancestor&_&&MethodName <ObjectParm>,<P1>,<P2>,<P3>,<P4>
      ENDM
    ENDIF
  ENDM

  ;; This macro declares a virtual method as external if the
  ;; current object is external.
  DECLARE_VMETHOD_EXTRN MACRO MethodName
    ;; see if the object is defined as external
    IFNB <&ExternalFlag>
      $DECL_EXTRN_ROUT _&ObjName&_&&MethodName&& ;; declare method as extrn
    ELSE
      IF $IsIncluded
	$DECL_EXTRN_ROUT _&ObjName&_&&MethodName&& ;; declare method as extrn
      ELSE
	PUBLIC _&ObjName&_&&MethodName	;; if not external, declare public
      ENDIF
    ENDIF
  ENDM

  ;; This macro makes a method invocation for the current
  ;; virtual method in the current object. If the object is
  ;; declared external, all methods will be as well, if not,
  ;; each method can individually be defined externally.
  MAKE_CURRENT_VMETHOD MACRO MethodName,MethodNum,VirtIndex

    ;; If the name and index have not been inherited, store them
    $&ObjName&_M&&MethodNum&&_Nm EQU <&&MethodName> ;; can access name by #
    $&ObjName&_VM&&MethodNum	EQU VirtIndex ;; store the virt method idx #

    ;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
    ObjName&_&&MethodName MACRO ObjectParm,P1,P2,P3,P4
      $GET_OBJ_REG_PARMS <ObjectParm>,<P1>,<P2>,<P3>,<P4>
      mov si,[di].$ODTOfs		;; get pointer to ODT
      $CALLPTR <DGROUP:[si].((VirtIndex SHL (CodePtrSize SHR 1))-(CodePtrSize-2))>
    ENDM
  ENDM

  ;; This macro makes a method invocation for the current
  ;; non-virtual method in the current object
  MAKE_CURRENT_CMETHOD MACRO MethodName,MethodNum,InlineFlag

    ;; If the name and index have not been inherited, store them
    $&ObjName&_M&&MethodNum&&_Nm EQU <&&MethodName> ;; can access name by #

    ;; if the method is not defined as external, an include file, or inline
    ;; make it public
    IFB <&&InlineFlag>
      IFB <&ExternalFlag>
	IFE $IsIncluded
	  PUBLIC _&ObjName&_&&MethodName&&
	ENDIF
      ENDIF
    ENDIF

    ;; see if the object is defined as external or an include file
    IFB <&&InlineFlag>
      IFNB <&ExternalFlag>
	;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
	ObjName&_&&MethodName MACRO ObjectParm,P1,P2,P3,P4
	  $DECL_EXTRN_ROUT _&ObjName&_&&MethodName&&
	  $GET_OBJ_REG_PARMS <ObjectParm>,<P1>,<P2>,<P3>,<P4>
	  call _&ObjName&_&&MethodName
	ENDM
      ELSE IF $IsIncluded
	;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
	ObjName&_&&MethodName MACRO ObjectParm,P1,P2,P3,P4
	  $DECL_EXTRN_ROUT _&ObjName&_&&MethodName&&
	  $GET_OBJ_REG_PARMS <ObjectParm>,<P1>,<P2>,<P3>,<P4>
	  call _&ObjName&_&&MethodName
	ENDM
      ELSE
	;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
	ObjName&_&&MethodName MACRO ObjectParm,P1,P2,P3,P4
	  $GET_OBJ_REG_PARMS <ObjectParm>,<P1>,<P2>,<P3>,<P4>
	  call _&ObjName&_&&MethodName
	ENDM
      ENDIF
    ELSE
      ;; define the macro, pass up to 4 register parameters for ax,bx,cx,dx
      ObjName&_&&MethodName MACRO ObjectParm,P1,P2,P3,P4
      ENDM
    ENDIF
  ENDM

  ;; This macro inherits a method from the ancestor to the
  ;; current object. It retains the name of the method, but prefaces
  ;; it by the current object name. In addition, it creates an ancestor
  ;; method with the same name, but followed by _A
  $INHERIT_ANC_METHOD MACRO CurrentMNum,MethodName
    IFDEF $&&MethodName&&_Ofs		;; if virtual, inherit virtual
      MAKE_ANCESTOR_METHOD &&MethodName,%$&Ancestor&_VM&&CurrentMNum
      MAKE_CURRENT_VMETHOD &&MethodName,&&CurrentMNum,%$&Ancestor&_VM&&CurrentMNum
    ELSE
      MAKE_ANCESTOR_METHOD &&MethodName ;; else inherit constant

      $&ObjName&_M&&CurrentMNum&&_Nm EQU <&&MethodName> ;; can access name by #

      ;; define as our macro, pass 4 standard parameters for ax,bx,cx,dx
      ObjName&_&&MethodName MACRO ObjectParm,P1,P2,P3,P4
	Ancestor&_&&MethodName <ObjectParm>,<P1>,<P2>,<P3>,<P4>
      ENDM
    ENDIF
  ENDM
  INHERIT_ANC_METHOD MACRO CurrentMNum
    $INHERIT_ANC_METHOD &&CurrentMNum,%$&Ancestor&_M&&CurrentMNum&&_Nm
  ENDM

  ;; This macro returns whether or not an ancestor method of
  ;; a certain name is defined
  IS_ANCESTOR_MNAME_DEF MACRO MethodName,TruthVar
    IFNB <Ancestor>
      IFDEF Ancestor&_&&MethodName
	TruthVar = TRUE
      ELSE
	TruthVar = FALSE
      ENDIF
    ELSE
      TruthVar = FALSE
    ENDIF
  ENDM

  ;; This macro returns whether or not an ancestor method of
  ;; a certain name is defined
  IS_ANCESTOR_MNUM_DEF MACRO MethodNum,TruthVar
    IFNB <Ancestor>
      IFDEF $&Ancestor&_M&&MethodNum&&_Nm
	TruthVar = TRUE
      ELSE
	TruthVar = FALSE
      ENDIF
    ELSE
      TruthVar = FALSE
    ENDIF
  ENDM

  ;; This macro makes an ODT entry with the offset of the passed
  ;; method routine.
  MAKE_CURRENT_ODT_ENTRY MACRO MName,CurVM
    $_&ObjName&_&&CurVM EQU @CODEOFS _&ObjName&_&&MName&&
  ENDM

  ;; If we don't redefine an ancestor method, this keeps the
  ;; continuity between objects.
  MAKE_ANC_ODT_EQU MACRO CurVM
    IFNDEF $_&ObjName&_&&CurVM
      $_&ObjName&_&&CurVM EQU $_&Ancestor&_&&CurVM ;; store ofs of ancestor
    ENDIF
  ENDM

  ;; This macro creates a symbol which can be accessed by the
  ;; the original virtual method number which holds
  ;; the offset of the current object's current method.
  ;; It also creates a symbol with the current
  ;; virtual method number which contains the offset of the
  ;; method which used to be at the new method's location.
  $MAKE_ANC_RELOC_ENTRY MACRO MName,CurVM,OldVM
    $_&ObjName&_&&CurVM EQU $_&Ancestor&_&&OldVM ;; store ofs of ancestor
    $_&ObjName&_&&OldVM EQU @CODEOFS _&ObjName&_&&MName&& ;; store new offset
  ENDM
  MAKE_ANC_RELOC_ENTRY MACRO MName,CurVM
    $MAKE_ANC_RELOC_ENTRY &&MName,&&CurVM,<%(($&&MName&&_Ofs+(CodePtrSize-2)) SHR (CodePtrSize SHR 1))>
  ENDM

  ;; This macro creates a symbol with the current virtual method
  ;; number which holds the offset of the new method defined.
  DECLARE_ODT_ENTRY MACRO CurVM
    ;; If this virtual method does not have a defined offset
    ;; associated with it, then it was inherited and not
    ;; replaced. We must define it so we can pass the value on
    ;; and declare a routine pointer for it in the currently open
    ;; segment.
    IFNDEF $_&ObjName&_&&CurVM
      $_&ObjName&_&&CurVM EQU $_&Ancestor&_&&CurVM ;; same as ancestor
    ENDIF
    dw $_&ObjName&_&&CurVM		;; declare the offset
    IF (CodePtrSize EQ 4)
      dw @CODE				;; if large model, put the segment
    ENDIF
  ENDM

  ;; This macro makes a data item declarator to preserve the
  ;; pre-initialization inheritance of data from object to
  ;; object
  MAKE_DATA_EQU MACRO Declare,CurDNum
    $&ObjName&_DDecl&&CurDNum EQU <&&Declare> ;; store in an equate
  ENDM

  ;; This macro makes a macro which can be used by the child
  ;; of this object to inherit all preinitialized data
  MAKE_ANCEST_DD MACRO NumDItems
    IF NumDItems
      $DO_DECL MACRO Declare		;; this macro does the declaration
	&&Declare
      ENDM
      IFDEF $&Ancestor&_DD		;; if ancestor's macro defined, do 1st
	$&ObjName&_DD MACRO Num
	  IFB <Num>
	    $&Ancestor&_DD		;; declare ancestor's data first
	    $&ObjName&_DD 1		;; start recursive loop
	  ELSE
	    $DO_DECL %$&ObjName&_DDecl&&&Num ;; declare this data item
	    IF (Num LT &&NumDItems)	;; do till we've declared all items
	      $&ObjName&_DD %(Num+1)	;; call recursively till done
	    ENDIF
	  ENDIF
	ENDM
      ELSE
	$&ObjName&_DD MACRO Num
	  IFB <Num>
	    $&ObjName&_DD 1		;; start loop, no ancestor data
	  ELSE
	    $DO_DECL %$&ObjName&_DDecl&&&Num ;; declare this data item
	    IF (Num LT &&NumDItems)	;; do till we've decalred all items
	      $&ObjName&_DD %(Num+1)	;; call recursively till done
	    ENDIF
	  ENDIF
	ENDM
      ENDIF
    ELSE
      IFDEF $&Ancestor&_DD		;; if ancestor's macro defined, do it
	$&ObjName&_DD MACRO
	  $&Ancestor&_DD		;; done
	ENDM
      ENDIF
    ENDIF
  ENDM

  ;; These macros declare data values, using these macros
  ;; helps to automate making object data public or private
  ODX MACRO LabelName,Declare,Dat,PublicFlag
    IFNB <&&PublicFlag>			;; if public specified, declare data
      &&LabelName &&Declare &&Dat	;; done
    ELSE				;; if not public, depends if we're extrn
      IFNB <&ExternalFlag>
	&&Declare &&Dat			;; if so, no label
      ELSE IF $IsIncluded		;; if we're included, we're extern
	&&Declare &&Dat			;; if so, no label
      ELSE
	&&LabelName &&Declare &&Dat	;; if not extrn, make data accessible
      ENDIF
    ENDIF
    ;; store data declarations for children to inherit
    $CurDataItem = $CurDataItem+1	;; add to the number of data items
    MAKE_DATA_EQU <&&Declare &&Dat>,%$CurDataItem ;; store the declaration
  ENDM
  ODB MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,db,<&&Dat>,&&PublicFlag
  ENDM
  ODW MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,dw,<&&Dat>,&&PublicFlag
  ENDM
  ODD MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,dd,<&&Dat>,&&PublicFlag
  ENDM
  ODF MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,df,<&&Dat>,&&PublicFlag
  ENDM
  ODP MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,dp,<&&Dat>,&&PublicFlag
  ENDM
  ODQ MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,dq,<&&Dat>,&&PublicFlag
  ENDM
  ODT MACRO LabelName,Dat,PublicFlag
    ODX &&LabelName,dt,<&&Dat>,&&PublicFlag
  ENDM

  ;; This macro lets you create an automatic initialization
  ;; and exit method for an object. The routine corresponding
  ;; to each name follows the normal method naming conventions.
  INIT_EXIT	  MACRO InitName,ExitName
    METHOD InitName			;; each will be a method
    METHOD ExitName			;; done
    IFE $IsIncluded
      IXMID	  SEGMENT
	  IF (CodePtrSize LT 4)
	    dw OFFSET _&ObjName&_&&InitName ;; first put the init routine offset
	    dw OFFSET _&ObjName&_&&ExitName ;; then the exit offset
	  ELSE
	    dw (OFFSET _&ObjName&_&&InitName),@CODE ;; put init routine pointer
	    dw (OFFSET _&ObjName&_&&ExitName),@CODE ;; then the exit
	  ENDIF
      IXMID	  ENDS
    ENDIF
  ENDM

  ;; This is used to define a method, if the method should be
  ;; virtual, the VirtualFlag flag should not be empty, if it will
  ;; be external, the ExternalFlag should not be empty.
  METHOD	  MACRO MethodName,VirtualFlag,InlineFlag
    IF $CurStructOpen			;; methods are declared after the data
      &ObjName ENDS			;; close the current structure
      $CurStructOpen = FALSE
    ENDIF

    ;; increment current method number
    $CurMethodNum = $CurMethodNum+1

    ;; virtual and non-virtual methods operate quite differently
    IFNB <VirtualFlag>

      $CurNumVirtMs = $CurNumVirtMs+1	;; increment number of virtual methods

      ;; see if the ancestor had this method as well
      IS_ANCESTOR_MNAME_DEF &&MethodName,$TempMethodDefined
      IF $TempMethodDefined

	IFNDEF $&&MethodName&&_Ofs	;; this will be defined for virt methods
	  .err				;; if not, trying to redefine non-virt
	  %out Error: redefinition of non-virtual method to virtual
	  EXITM
	ENDIF

	;; make a method which will call the ancestor
	;; we have already inherited the proper calling macro for
	;; our method
	MAKE_ANCESTOR_METHOD &&MethodName,%$CurNumVirtMs

	;; Also make a relocation entry for the original virtual
	;; method entry index
	MAKE_ANC_RELOC_ENTRY &&MethodName,%$CurNumVirtMs

      ELSE

	;; since this is the first definition of this method,
	;; make the original method offset
	IF (CodePtrSize EQ 2)
	  $&&MethodName&&_Ofs EQU ($CurNumVirtMs SHL 1)
	ELSE
	  $&&MethodName&&_Ofs EQU (($CurNumVirtMs SHL 2)-2)
	ENDIF

	;; Make a method macro for this method
	MAKE_CURRENT_VMETHOD &&MethodName,%$CurMethodNum,%$CurNumVirtMs

	;; Make an ODT entry with the offset of this routine
	MAKE_CURRENT_ODT_ENTRY &&MethodName,%$CurNumVirtMs

      ENDIF

      ;; declare this routine as external if it should be
      DECLARE_VMETHOD_EXTRN &&MethodName
    ELSE

      IFDEF $&&MethodName&&_Ofs
	;; if redefined as not virtual, error
	.err
	%out Error: redefinition of virtual method &&MethodName as non-virtual
	EXITM
      ENDIF

      MAKE_CURRENT_CMETHOD &&MethodName,%$CurMethodNum,InlineFlag

    ENDIF
  ENDM

  ;; This must be used to close an object definition
  ;; we need to define the virtual method table definition and
  ;; finalize last offsets, etc. Make the macro for <ObjName>
  OBJECT_END	  MACRO

    IF $CurStructOpen
      &ObjName ENDS			;; close the current structure
      $CurStructOpen = FALSE
    ENDIF

    $CurObjectOpen = FALSE		;; object not open anymore

    $&ObjName&_LastMet = $CurMethodNum	;; store the number of our last method
    $&ObjName&_VMs = $CurNumVirtMs	;; store the number of virtual methods
    MAKE_ANCEST_DD %$CurDataItem	;; make the data inheritance macro

    $CurNumVirtMs = 0			;; start with 0 again
    REPT $&ObjName&_VMs
      $CurNumVirtMs = $CurNumVirtMs+1	;; loop through VMs
      MAKE_ANC_ODT_EQU %$CurNumVirtMs	;; make the equates that aren't made yet
    ENDM

    ;; Start the data segment to put the ODT
    @SEG_START %@DATA
      ;; If the ODT for this object is not external, define it.
      IFB <ExternalFlag>
	IFE $IsIncluded
	  dw (SIZE &ObjName)		;; first entry is the size

	  $CurNumVirtMs = 0		;; reset our counter to 0
	  REPT $&ObjName&_VMs		;; go through all virtual methods
	    $CurNumVirtMs = $CurNumVirtMs+1 ;; reset our counter to 0
	    DECLARE_ODT_ENTRY %$CurNumVirtMs ;; piece together ODT
	  ENDM
	ENDIF
      ENDIF
    @SEG_END %@DATA

    PURGE @SEG_START,@SEG_END
    PURGE MAKE_ANCESTOR_METHOD,DECLARE_VMETHOD_EXTRN,MAKE_CURRENT_VMETHOD
    PURGE MAKE_CURRENT_CMETHOD,$INHERIT_ANC_METHOD,INHERIT_ANC_METHOD
    PURGE $MAKE_ANC_RELOC_ENTRY,MAKE_ANC_RELOC_ENTRY,IS_ANCESTOR_MNAME_DEF
    PURGE IS_ANCESTOR_MNUM_DEF,MAKE_CURRENT_ODT_ENTRY,DECLARE_ODT_ENTRY
    PURGE MAKE_ANC_ODT_EQU,MAKE_DATA_EQU,MAKE_ANCEST_DD
    PURGE ODB,ODW,ODD,ODF,ODP,ODQ,ODT,ODX,METHOD,INIT_EXIT,OBJECT_END
  ENDM

  ;;------------------------------------------------------------
  ;;
  ;; THE ACTUAL BODY OF THE OBJECT MACRO STARTS HERE
  ;;
  ;;------------------------------------------------------------
  ;; Start the data segment to put the ODT
  @SEG_START %@DATA
    ;; If the ODT for this object is not external, define it.
    IFNB <ExternalFlag>
      EXTRN $&ObjName&_ODT:Word		;; define the ODT as external
    ELSE IF $IsIncluded
      EXTRN $&ObjName&_ODT:Word		;; define the ODT as external
    ELSE
      PUBLIC $&ObjName&_ODT		;; make ODT public so it can be external
      $&ObjName&_ODT LABEL Word		;; the start of the ODT is size (init=0)
    ENDIF
  @SEG_END %@DATA


  IFB <Ancestor>
    $CurNumVirtMs = 0			;; number of virtual methods in object
    $CurMethodNum = 0			;; number of the current method
  ELSE
    ;; start virtual methods after the ancestor's last
    $CurNumVirtMs = $&Ancestor&_VMs	;; start with same number of methods

    ;; This loop inherits all of the ancestor's method definitions
    $CurMethodNum  = 0			;; keep track of total number of methods
    REPT      $&Ancestor&_LastMet	;; define our virtual methods as ancest
      $CurMethodNum = $CurMethodNum+1	;; pre-increment so we end with current

      ;; See if this number has a method associated with it
      IS_ANCESTOR_MNUM_DEF %$CurMethodNum,$TempMethodDefined

      IF $TempMethodDefined		;; if so, inherit it
	INHERIT_ANC_METHOD %$CurMethodNum ;; done
      ENDIF
    ENDM
  ENDIF

  ;; an object is a structure +
  ObjName     STRUC
  dw	  @DATAOFS $&ObjName&_ODT	;; first element is always $ODTOfs

  IFDEF $&Ancestor&_DD			;; if macro defined to inherit data
    $&Ancestor&_DD			;; declare ancestor's data
  ENDIF

ENDM

; This macro includes an object assembly source file and
; makes all of the objects inside it external. It also references the
; OBJECTS.LIB library file.
USE_OBJECTS MACRO FNames
  INCLUDELIB OBJECTS.LIB		;; include the objects library
  IRP X,<FNames>
    IF $IsIncluded			;; if already included, don't change
      INCLUDE &X&.O			;; include the object file .O extension
    ELSE
      $IsIncluded = TRUE		;; specify that this is included
      INCLUDE &X&.O			;; include the object file .O extension
      $IsIncluded = FALSE		;; not included anymore
    ENDIF
  ENDM
ENDM

STERILIZE MACRO ObjList

  ;; This macro removes the memory needed by an object
  ;; in order to produce offspring. After sterilizing an
  ;; object, you cannot inherit from it.
  STERIL1 MACRO ObjName
    IFDEF $&&ObjName&&_DD
      PURGE $&&ObjName&&_DD		;; remove the data declaration macro
    ENDIF

    REDEFEQU MACRO MetNum
      $&&ObjName&&_M&&&MetNum&&&_Nm EQU <> ;; release memory used for this equate
      IFDEF $_&&ObjName&&_&&&MetNum&&&
	$_&&ObjName&&_&&&MetNum&&& EQU <> ;; and this one
      ENDIF
    ENDM

    $CurMethodNum  = 0			;; keep track of current method
    REPT $&&ObjName&&_LastMet
      $CurMethodNum = $CurMethodNum+1	;; pre-increment so we end with current
      REDEFEQU %$CurMethodNum		;; release method equate
    ENDM

    PURGE REDEFEQU
  ENDM

  ;; Sterilize all objects in the list
  IRP X,<ObjList>
    STERIL1 X
  ENDM

  PURGE STERIL1
ENDM
